@@ -4,7 +4,6 @@ from .filterable import Filterable
 from .hashable import Hashable
 from .identifiable import Identifiable
 from .labelable import Labelable
-from .pathable import Pathable
 from .serialisable import Serialisable
 from .taggable import Taggable
 
@@ -16,7 +15,6 @@ __all__ = (
     "Identifiable",
     "Labelable",
     "MixinBase",
-    "Pathable",
     "Serialisable",
     "Taggable",
 )
