@@ -12,6 +12,9 @@ import google.api_core.exceptions
 import pkg_resources
 from google_crc32c import Checksum
 
+from octue.resources.label import LabelSet
+from octue.resources.tag import TagDict
+
 
 try:
     import h5py
@@ -20,55 +23,36 @@ except ModuleNotFoundError:
 
 from octue.cloud import storage
 from octue.cloud.storage import GoogleCloudStorageClient
-from octue.exceptions import CloudLocationNotSpecified, FileNotFoundException, InvalidInputException
+from octue.exceptions import CloudLocationNotSpecified
 from octue.migrations.cloud_storage import translate_bucket_name_and_path_in_bucket_to_cloud_path
-from octue.mixins import Filterable, Hashable, Identifiable, Labelable, Pathable, Serialisable, Taggable
+from octue.mixins import Filterable, Hashable, Identifiable, Labelable, Serialisable, Taggable
 from octue.mixins.hashable import EMPTY_STRING_HASH_VALUE
-from octue.utils import isfile
+from octue.utils.decoders import OctueJSONDecoder
 from octue.utils.encoders import OctueJSONEncoder
-from octue.utils.local_metadata import LOCAL_METADATA_FILENAME, load_local_metadata_file
+from octue.utils.metadata import METADATA_FILENAME, load_local_metadata_file
 
 
 logger = logging.getLogger(__name__)
 
+
 OCTUE_METADATA_NAMESPACE = "octue"
 
-ID_DEFAULT = None
-TAGS_DEFAULT = None
-LABELS_DEFAULT = None
 
+class Datafile(Labelable, Taggable, Serialisable, Identifiable, Hashable, Filterable):
+    """A representation of a data file on the Octue system. Metadata for the file is obtained from its corresponding
+    cloud object or a local `.octue` metadata file, if present. If no stored metadata is available, it can be set during
+    or after instantiation.
 
-class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashable, Filterable):
-    """A representation of a data file on the Octue system. If the given path is a cloud path and `hypothetical` is not
-    `True`, the datafile's metadata is pulled from the given cloud location, and any conflicting parameters (see the
-    `Datafile.metadata` method description for the parameter names concerned) are ignored. The metadata of cloud
-    datafiles can be changed using the `Datafile.update_metadata` method, but not during instantiation.
-
-    Files in a manifest look like this:
-
-        {
-          "path": "folder/subfolder/file_1.csv",
-          "extension": "csv",
-          "tags": {},
-          "labels": [],
-          "timestamp": datetime.datetime(2021, 5, 3, 18, 15, 58, 298086),
-          "id": "abff07bc-7c19-4ed5-be6d-a6546eae8e86",
-          "size_bytes": 59684813,
-          "sha-512/256": "somesha"
-        },
-
-    :param str|None path: The path of this file locally or in the cloud, which may include folders or subfolders, within the dataset. If no path_from parameter is set, then absolute paths are acceptable, otherwise relative paths are required.
+    :param str|None path: The path of this file locally or in the cloud, which may include folders or subfolders, within the dataset
     :param str|None local_path: If a cloud path is given as the `path` parameter, this is the path to an existing local file that is known to be in sync with the cloud object
     :param str|None cloud_path: If a local path is given for the `path` parameter, this is a cloud path to keep in sync with the local file
     :param datetime.datetime|int|float|None timestamp: A posix timestamp associated with the file, in seconds since epoch, typically when it was created but could relate to a relevant time point for the data
     :param str id: The Universally Unique ID of this file (checked to be valid if not None, generated if None)
-    :param Pathable path_from: The root Pathable object (typically a Dataset) that this Datafile's path is relative to.
-    :param dict|TagDict tags: key-value pairs with string keys conforming to the Octue tag format (see TagDict)
-    :param iter(str) labels: Space-separated string of labels relevant to this file
-    :param bool skip_checks:
+    :param dict|octue.resources.tag.TagDict|None tags: key-value pairs with string keys conforming to the Octue tag format (see TagDict)
+    :param iter(str)|octue.resources.label.LabelSet|None labels: Space-separated string of labels relevant to this file
     :param str mode: if using as a context manager, open the datafile for reading/editing in this mode (the mode options are the same as for the builtin open function)
-    :param bool update_cloud_metadata: if using as a context manager and this is True, update the cloud metadata of the datafile when the context is exited
-    :param bool hypothetical: True if the file does not actually exist or access is not available at instantiation
+    :param bool update_cloud_metadata: if using as a context manager and this is `True`, update the cloud metadata of the datafile when the context is exited
+    :param bool hypothetical: if `True`, ignore any metadata stored for this datafile locally or in the cloud and use whatever is given at instantiation
     :return None:
     """
 
@@ -89,11 +73,9 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
         local_path=None,
         cloud_path=None,
         timestamp=None,
-        id=ID_DEFAULT,
-        path_from=None,
-        tags=TAGS_DEFAULT,
-        labels=LABELS_DEFAULT,
-        skip_checks=True,
+        id=None,
+        tags=None,
+        labels=None,
         mode="r",
         update_cloud_metadata=True,
         hypothetical=False,
@@ -105,71 +87,49 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
             immutable_hash_value=kwargs.pop("immutable_hash_value", None),
             tags=tags,
             labels=labels,
-            path=path,
-            path_from=path_from,
         )
 
         self.timestamp = timestamp
-
+        self._open_attributes = {"mode": mode, "update_cloud_metadata": update_cloud_metadata, **kwargs}
         self._local_path = None
         self._cloud_path = None
-        self._hypothetical = hypothetical
-        self._open_attributes = {"mode": mode, "update_cloud_metadata": update_cloud_metadata, **kwargs}
         self._cloud_metadata = {}
 
-        if storage.path.is_cloud_path(self.path):
-            self._cloud_path = path
-
-            if not self._hypothetical:
-                # Collect any non-`None` metadata instantiation parameters so the user can be warned if they conflict
-                # with any metadata already on the cloud object.
-                initialisation_parameters = {}
-
-                for parameter in ("id", "timestamp", "tags", "labels"):
-                    value = locals().get(parameter)
-                    if value is not None:
-                        initialisation_parameters[parameter] = value
-
-                self._use_cloud_metadata(**initialisation_parameters)
-
-            if local_path:
-                # If there is no file at the given local path or the file is different to the one in the cloud, download
-                # the cloud file locally.
-                if not os.path.exists(local_path) or self.cloud_hash_value != calculate_hash(local_path):
-                    self.download(local_path)
-                else:
-                    self._local_path = local_path
-
+        if storage.path.is_cloud_path(path):
+            self._instantiate_from_cloud_object(path, local_path, ignore_stored_metadata=hypothetical)
         else:
-            self._local_path = self.absolute_path
-            self._get_local_metadata()
+            self._instantiate_from_local_path(path, cloud_path, ignore_stored_metadata=hypothetical)
 
-            # Run integrity checks on the file.
-            if not skip_checks:
-                self.check(**kwargs)
-
-            if cloud_path:
-                self.cloud_path = cloud_path
+        if hypothetical:
+            logger.debug("Ignored stored metadata for %r.", self)
+        else:
+            if self.metadata(use_octue_namespace=False, include_sdk_version=False) != {
+                "id": id or self.id,
+                "timestamp": timestamp,
+                "tags": TagDict(tags),
+                "labels": LabelSet(labels),
+            }:
+                logger.warning(
+                    "Overriding metadata given at instantiation with stored metadata for %r - set `hypothetical` to "
+                    "`True` at instantiation to avoid this.",
+                    self,
+                )
 
     @classmethod
-    def deserialise(cls, serialised_datafile, path_from=None):
-        """Deserialise a Datafile from a dictionary. The `path_from` parameter is only used if the path in the
-        serialised Dataset is relative.
+    def deserialise(cls, serialised_datafile, from_string=False):
+        """Deserialise a Datafile from a dictionary or JSON string.
 
-        :param dict serialised_datafile:
-        :param octue.mixins.Pathable path_from:
+        :param dict|str serialised_datafile:
+        :param bool from_string:
         :return Datafile:
         """
-        serialised_datafile = copy.deepcopy(serialised_datafile)
-        cloud_metadata = serialised_datafile.pop("_cloud_metadata", {})
-
-        if not os.path.isabs(serialised_datafile["path"]) and not storage.path.is_qualified_cloud_path(
-            serialised_datafile["path"]
-        ):
-            datafile = Datafile(**serialised_datafile, path_from=path_from)
+        if from_string:
+            serialised_datafile = json.loads(serialised_datafile, cls=OctueJSONDecoder)
         else:
-            datafile = Datafile(**serialised_datafile)
+            serialised_datafile = copy.deepcopy(serialised_datafile)
 
+        cloud_metadata = serialised_datafile.pop("_cloud_metadata", {})
+        datafile = Datafile(**serialised_datafile)
         datafile._cloud_metadata = cloud_metadata
         return datafile
 
@@ -179,7 +139,13 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
 
         :return str:
         """
-        return self._name or str(os.path.split(self.path)[-1]).split("?")[0]
+        if self._name:
+            return self._name
+
+        if self.exists_in_cloud:
+            return str(storage.path.split(self.cloud_path)[-1]).split("?")[0]
+
+        return str(os.path.split(self.local_path)[-1])
 
     @property
     def extension(self):
@@ -295,9 +261,9 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
     def _last_modified(self):
         """Get the date/time the file was last modified in units of seconds since epoch (posix time).
 
-        :return float:
+        :return float|None:
         """
-        if self._path_is_in_google_cloud_storage:
+        if self.exists_in_cloud:
             last_modified = self._cloud_metadata.get("updated")
 
             if last_modified is None:
@@ -305,18 +271,24 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
 
             return last_modified.timestamp()
 
-        return os.path.getmtime(self.absolute_path)
+        try:
+            return os.path.getmtime(self.local_path)
+        except FileNotFoundError:
+            return None
 
     @property
     def size_bytes(self):
         """Get the size of the datafile in bytes.
 
-        :return float:
+        :return float|None:
         """
-        if self._path_is_in_google_cloud_storage:
+        if self.exists_in_cloud:
             return self._cloud_metadata.get("size")
 
-        return os.path.getsize(self.absolute_path)
+        try:
+            return os.path.getsize(self.local_path)
+        except FileNotFoundError:
+            return None
 
     @property
     def exists_in_cloud(self):
@@ -338,12 +310,22 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
     def local_path(self):
         """Get the local path for the datafile, downloading it from the cloud to a temporary file if necessary. If
         downloaded, the local path is added to a cache to avoid downloading again in the same runtime.
+
+        :return str:
         """
         if self._local_path:
             return self._local_path
 
         return self.download()
 
+    @property
+    def path(self):
+        """Alias to the `local_path` property.
+
+        :return str:
+        """
+        return self.local_path
+
     @local_path.setter
     def local_path(self, path):
         """Set the local path of the datafile and:
@@ -395,7 +377,7 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
         if not self.exists_locally:
             return None
 
-        return os.path.join(os.path.dirname(self._local_path), LOCAL_METADATA_FILENAME)
+        return os.path.join(os.path.dirname(self._local_path), METADATA_FILENAME)
 
     def __enter__(self):
         self._open_context_manager = self.open(**self._open_attributes)
@@ -403,16 +385,17 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         self._open_context_manager.__exit__(exc_type, exc_val, exc_tb)
+        del vars(self)["_open_context_manager"]
 
     def __lt__(self, other):
         if not isinstance(other, Datafile):
             raise TypeError(f"An object of type {type(self)} cannot be compared with {type(other)}.")
-        return self.absolute_path < other.absolute_path
+        return self.name < other.name
 
     def __gt__(self, other):
         if not isinstance(other, Datafile):
             raise TypeError(f"An object of type {type(self)} cannot be compared with {type(other)}.")
-        return self.absolute_path > other.absolute_path
+        return self.name > other.name
 
     def to_cloud(self, cloud_path=None, bucket_name=None, path_in_bucket=None, update_cloud_metadata=True):
         """Upload a datafile to Google Cloud Storage.
@@ -440,10 +423,10 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
 
         if update_cloud_metadata:
             # If the datafile's metadata has been changed locally, update the cloud file's metadata.
-            local_metadata = self.metadata()
+            local_metadata = self.metadata(use_octue_namespace=False)
 
             if self._cloud_metadata.get("custom_metadata") != local_metadata:
-                self._update_cloud_metadata()
+                self.update_cloud_metadata()
 
         return self.cloud_path
 
@@ -480,33 +463,12 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
         self.reset_hash()
         return self._local_path
 
-    def check(self, size_bytes=None, sha=None, last_modified=None, extension=None):
-        """Check file presence and integrity"""
-        # TODO Check consistency of size_bytes input against self.size_bytes property for a file if we have one
-        # TODO Check consistency of sha against file contents if we have a file
-        # TODO Check consistency of last_modified date
-
-        if (extension is not None) and not self.path.endswith(extension):
-            raise InvalidInputException(
-                f"Extension provided ({extension}) does not match file extension (from {self.path}). Pass extension="
-                f"None to set extension from filename automatically."
-            )
-
-        if not self.exists():
-            raise FileNotFoundException(f"No file found at {self.absolute_path}")
-
-    def exists(self):
-        """Return `True` if the datafile exists on the current system.
-
-        :return bool:
-        """
-        return isfile(self.absolute_path)
-
-    def metadata(self, use_octue_namespace=True):
+    def metadata(self, include_sdk_version=True, use_octue_namespace=True):
         """Get the datafile's metadata in a serialised form (i.e. the attributes `id`, `timestamp`, `labels`, `tags`,
         and `sdk_version`).
 
-        :param bool use_octue_namespace: if True, prefix metadata names with "octue__"
+        :param bool include_sdk_version: if `True`, include the `octue` version that instantiated the datafile in the metadata
+        :param bool use_octue_namespace: if `True`, prefix metadata names with "octue__"
         :return dict:
         """
         metadata = {
@@ -514,32 +476,17 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
             "timestamp": self.timestamp,
             "tags": self.tags,
             "labels": self.labels,
-            "sdk_version": pkg_resources.get_distribution("octue").version,
         }
 
+        if include_sdk_version:
+            metadata["sdk_version"] = pkg_resources.get_distribution("octue").version
+
         if not use_octue_namespace:
             return metadata
 
         return {f"{OCTUE_METADATA_NAMESPACE}__{key}": value for key, value in metadata.items()}
 
-    def _get_cloud_metadata(self):
-        """Get the cloud metadata for the datafile.
-
-        :return None:
-        """
-        if not self.cloud_path:
-            self._raise_cloud_location_error()
-
-        # Skip getting metadata for now if the cloud path is a signed URL.
-        if storage.path.is_url(self.cloud_path):
-            return
-
-        cloud_metadata = GoogleCloudStorageClient().get_metadata(cloud_path=self.cloud_path)
-
-        if cloud_metadata:
-            self._cloud_metadata = cloud_metadata
-
-    def _update_cloud_metadata(self):
+    def update_cloud_metadata(self):
         """Update the cloud metadata for the datafile.
 
         :return None:
@@ -549,55 +496,7 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
 
         GoogleCloudStorageClient().overwrite_custom_metadata(metadata=self.metadata(), cloud_path=self.cloud_path)
 
-    def _use_cloud_metadata(self, **initialisation_parameters):
-        """Populate the datafile's attributes from the metadata of the cloud object located at its path (by necessity a
-        cloud path) and project name. If there is a conflict between the cloud metadata and a given local initialisation
-        parameter, the local value is used.
-
-        :param initialisation_parameters: key-value pairs of initialisation parameter names and values (provide to check for conflicts with cloud metadata)
-        :return None:
-        """
-        self._get_cloud_metadata()
-        cloud_custom_metadata = self._cloud_metadata.get("custom_metadata", {})
-        self._warn_about_attribute_conflicts(cloud_custom_metadata, **initialisation_parameters)
-
-        self._set_id(
-            initialisation_parameters.get(
-                "id", cloud_custom_metadata.get(f"{OCTUE_METADATA_NAMESPACE}__id", ID_DEFAULT)
-            )
-        )
-
-        self.immutable_hash_value = self.cloud_hash_value or EMPTY_STRING_HASH_VALUE
-
-        for attribute in ("timestamp", "tags", "labels"):
-            setattr(
-                self,
-                attribute,
-                initialisation_parameters.get(
-                    attribute, cloud_custom_metadata.get(f"{OCTUE_METADATA_NAMESPACE}__{attribute}")
-                ),
-            )
-
-    def _get_local_metadata(self):
-        """Get the datafile's local metadata from the local metadata records file and apply it to the datafile instance.
-        If no metadata is stored for the datafile, do nothing.
-
-        :return None:
-        """
-        existing_metadata_records = load_local_metadata_file(self._local_metadata_path)
-        datafile_metadata = existing_metadata_records.get("datafiles", {}).get(self.name, {})
-
-        if not datafile_metadata:
-            return
-
-        if "id" in datafile_metadata:
-            self._set_id(datafile_metadata["id"])
-
-        for parameter in ("timestamp", "tags", "labels"):
-            if parameter in datafile_metadata:
-                setattr(self, parameter, datafile_metadata[parameter])
-
-    def _update_local_metadata(self):
+    def update_local_metadata(self):
         """Create or update the local octue metadata file with the datafile's metadata.
 
         :return None:
@@ -612,25 +511,106 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashab
         with open(self._local_metadata_path, "w") as f:
             json.dump(existing_metadata_records, f, cls=OctueJSONEncoder)
 
-    def _warn_about_attribute_conflicts(self, cloud_custom_metadata, **initialisation_parameters):
-        """Raise a warning if there is a conflict between the cloud custom metadata and the given initialisation
-        parameters if the cloud value is not `None` or an empty collection.
+    def _instantiate_from_cloud_object(self, path, local_path, ignore_stored_metadata):
+        """Instantiate the datafile from a cloud object.
 
-        :param dict cloud_custom_metadata:
+        :param str path: the cloud path to a cloud object
+        :param str|None local_path: a local path to use for opening or modifying the cloud object locally
+        :param bool ignore_stored_metadata: if `True`, don't use any metadata stored for this datafile in the cloud
         :return None:
         """
-        for attribute_name, attribute_value in initialisation_parameters.items():
+        self._cloud_path = path
 
-            cloud_metadata_value = cloud_custom_metadata.get(f"{OCTUE_METADATA_NAMESPACE}__{attribute_name}")
+        if not ignore_stored_metadata:
+            self._use_cloud_metadata()
 
-            if cloud_metadata_value == attribute_value or not cloud_metadata_value:
-                continue
+        if local_path:
+            # If there is no file at the given local path or the file is different to the one in the cloud, download
+            # the cloud file locally.
+            if not os.path.exists(local_path) or self.cloud_hash_value != calculate_hash(local_path):
+                self.download(local_path)
+            else:
+                self._local_path = local_path
 
-            logger.warning(
-                f"The value {cloud_metadata_value!r} of the {type(self).__name__} attribute {attribute_name!r} from "
-                f"the cloud conflicts with the value given locally at instantiation {attribute_value!r}. The local "
-                f"value has been used and will overwrite the cloud value if the datafile is saved."
-            )
+    def _instantiate_from_local_path(self, path, cloud_path, ignore_stored_metadata):
+        """Instantiate the datafile from a local path.
+
+        :param str path: the path to a local file
+        :param str|None cloud_path: a cloud path to upload the datafile to and mirror any changes made to it locally
+        :param bool ignore_stored_metadata: if `True`, don't use any metadata stored for this datafile locally
+        :return None:
+        """
+        self._local_path = os.path.abspath(path)
+
+        if not ignore_stored_metadata:
+            self._use_local_metadata()
+
+        if cloud_path:
+            self.cloud_path = cloud_path
+
+    def _get_cloud_metadata(self):
+        """Get the cloud metadata for the datafile and store it without updating the datafile's metadata.
+
+        :return None:
+        """
+        if not self.cloud_path:
+            self._raise_cloud_location_error()
+
+        # Skip getting metadata for now if the cloud path is a signed URL.
+        if storage.path.is_url(self.cloud_path):
+            return
+
+        cloud_metadata = GoogleCloudStorageClient().get_metadata(cloud_path=self.cloud_path)
+
+        if cloud_metadata:
+            cloud_metadata["custom_metadata"] = {
+                key.replace(f"{OCTUE_METADATA_NAMESPACE}__", ""): value
+                for key, value in cloud_metadata["custom_metadata"].items()
+            }
+
+            self._cloud_metadata = cloud_metadata
+
+    def _use_cloud_metadata(self):
+        """Update the datafile's metadata from the metadata of the cloud object located at its path. If no metadata is
+        stored for the datafile, do nothing.
+
+        :return None:
+        """
+        self._get_cloud_metadata()
+        cloud_custom_metadata = self._cloud_metadata.get("custom_metadata", {})
+
+        if not cloud_custom_metadata:
+            return
+
+        self.immutable_hash_value = self.cloud_hash_value or EMPTY_STRING_HASH_VALUE
+        self._set_metadata(cloud_custom_metadata)
+
+    def _use_local_metadata(self):
+        """Update the datafile's metadata from the local metadata records file. If no metadata is stored for the
+        datafile, do nothing.
+
+        :return None:
+        """
+        existing_metadata_records = load_local_metadata_file(self._local_metadata_path)
+        datafile_metadata = existing_metadata_records.get("datafiles", {}).get(self.name, {})
+
+        if not datafile_metadata:
+            return
+
+        self._set_metadata(datafile_metadata)
+
+    def _set_metadata(self, metadata):
+        """Set the instance's metadata.
+
+        :param dict metadata:
+        :return None:
+        """
+        if "id" in metadata:
+            self._set_id(metadata["id"])
+
+        for attribute in ("timestamp", "tags", "labels"):
+            if attribute in metadata:
+                setattr(self, attribute, metadata[attribute])
 
     def _calculate_hash(self):
         """Get the hash of the datafile according to the first of the following methods that is applicable:
@@ -692,8 +672,8 @@ class _DatafileContextManager:
     This is equivalent to the standard python:
     ```
     my_datafile = Datafile(path='subfolder/subsubfolder/my_datafile.json)
-    os.makedirs(os.path.split(my_datafile.absolute_path)[0], exist_ok=True)
-    with open(my_datafile.absolute_path, 'w') as fp:
+    os.makedirs(os.path.split(my_datafile.local_path)[0], exist_ok=True)
+    with open(my_datafile.local_path, 'w') as fp:
         fp.write("{}")
     ```
 
@@ -746,10 +726,9 @@ class _DatafileContextManager:
         if any(character in self.mode for character in self.MODIFICATION_MODES):
 
             # If the datafile is local-first, update its local metadata.
-            if not storage.path.is_qualified_cloud_path(self.datafile.path):
-                self.datafile._update_local_metadata()
-
-            if self.datafile.exists_in_cloud:
+            if not self.datafile.exists_in_cloud:
+                self.datafile.update_local_metadata()
+            else:
                 self.datafile.to_cloud(update_cloud_metadata=self._update_cloud_metadata)
 
 
